import os
import logging


from time import time
from typing import Callable

from fastapi import  Response, Request, HTTPException
from fastapi.routing import APIRoute

from .loggers import TNLogger, UnExpectedError


class FastAPILogger(APIRoute):
        

    def get_route_handler(self)  -> Callable:
        original_route_handler = super().get_route_handler()

        # resource = TNLogger.with_service_detail()
        error_logger = TNLogger.with_service_detail(name="error", level=logging.ERROR)

        # uvicorn_access = logging.getLogger("uvicorn.access")
        # uvicorn_access.addHandler(fast_logger.get_handler())

        async def custom_route_handler(request: Request) -> Response:
            
            try:
                response: Response = await original_route_handler(request)
                return response
            except HTTPException as http_exc:
                error_logger.error('{}'.format(http_exc.detail))
                raise HTTPException(http_exc.status_code, detail=str(http_exc.detail))

            except Exception as exc:
                obj_err = UnExpectedError("NO Handler Error", exc=exc)

                if obj_err.http_status == 500:
                    error_logger.error(obj_err.msg_body, stack_info=True, exc_info=True, extra=obj_err.attributes)
                else:
                    error_logger.error(obj_err.msg_body, extra=obj_err.attributes)
                    
                raise HTTPException(500, detail = str(exc))


        return custom_route_handler


import os

from pathlib import Path
from os import environ


def clear_multiproc_dir():
    """
    This function will clear all files in PROMETHEUS_MULTIPROC_DIR

    This function need to be call everytime before execute service that use multiprocess(have multiple workers)
    """
    MULTIPROC_DIR = environ.get("PROMETHEUS_MULTIPROC_DIR")
    if MULTIPROC_DIR:
        file_path = Path(MULTIPROC_DIR)
        if os.path.isdir(file_path):
            for f in os.listdir(file_path):
                os.remove(os.path.join(file_path, f))
        else:
            os.mkdir(file_path)



from opentelemetry.sdk.resources import Resource

def get_resource_info(resource: Resource):
    if resource:
        return resource.attributes['service.namespace'], resource.attributes['app.name']
    else:
        return 'unknown_namespace', 'unknown_service_name'


import time

from datetime import datetime


start_time = time.time()

def get_healthy():
    uptime_ms = (time.time() - start_time) * 1000
    response = {
        "uptime": round(uptime_ms, 2),
        "timestamp": str(datetime.now()),
        "connection": {"HTTPServer": True, "mongo": True},
    }
    status = 200

    return response


def get_healthz():
    uptime_ms = (time.time() - start_time) * 1000
    response = {
        "uptime": round(uptime_ms, 2),
        "timestamp": str(datetime.now()),
        "connection": {"HTTPServer": True},
    }
    status = 200

    return response




import sys
import numpy as np


class Calculator:
    """
    Instantiate a calulator.
    """

    def __init__(self):
        self.previous_result = []
        self.previous_operation = []
        self.previous_input = []
        pass

    def add(self, number1, number2):
        """
        add 2 given number.
        
        :param number1,number2: The numbers to add.
        :type number1,number2: float
    
        :return: The result of the addition.
        :rtype: float
        """
        input = [number1, number2]
        result = np.add(number1, number2)
        self._set_history(input, result)

        return result

    def multiply(self, number1, number2):
        """
        Multiply 2 given number.
        
        :param number1,number2: The numbers to multiply.
        :type number1,number2: float
    
        :return: The result of the multiplication.
        :rtype: float
        """
        input = [number1, number2]
        result = np.dot(number1, number2)
        self._set_history(input, result)

        return result

    def _set_history(self, input, result):
        if len(self.previous_result) > 10:
            self.previous_result.pop(0)
            self.previous_operation.pop(0)
            self.previous_input.pop(0)

        operation = sys._getframe().f_back.f_code.co_name

        self.previous_result.append(result)
        self.previous_operation.append(operation)
        self.previous_input.append(input)

    def get_history(self):
        """
        Get history of this calculator operation.

        """
        if len(self.previous_result) == 0:
            return None

        dict_map = {"add": "+", "multiply": "*"}

        for i in range(len(self.previous_result)):
            print(
                f"{i} ::\t{self.previous_input[i][0]}\t{dict_map[self.previous_operation[i]]}\t{self.previous_input[i][1]}\t=\t{self.previous_result[i]}\n"
            )




from .config import _DEFAULT_BUCKETS
from .config import NAMESPACE, SERVICENAME, DICT_REQURIED_LABELS


def add_default_label_value(func):
    """Decorator that add default label to the custom metrics"""

    def wrap(*args, **kwargs):
        if kwargs:
            for key in DICT_REQURIED_LABELS:
                if not kwargs.get(key):
                    kwargs.update({key: DICT_REQURIED_LABELS[key]})
        elif args:
            list_args = list(args)
            for requried_label in DICT_REQURIED_LABELS.keys():
                if requried_label not in list_args:
                    list_args.append(DICT_REQURIED_LABELS[requried_label])
            args = tuple(list_args)

        result = func(*args, **kwargs)

        return result

    return wrap


class CustomMetricsWrapperBase(object):
    """
    This Class use to warp Metrics class so that it will be include default label's value autometiclly when observed/used
    """

    def __init__(self, obj):
        self._wrapped_obj = obj

    @add_default_label_value
    def labels(self, *args, **kwargs):
        if kwargs:
            return self._wrapped_obj.labels(**kwargs)
        elif args:
            return self._wrapped_obj.labels(*args)

    def clear(self):
        self._wrapped_obj.clear()


class CounterWrapper(CustomMetricsWrapperBase):
    def __init__(self, obj):
        super().__init__(obj)

    def inc(self, *args):
        """Increment counter by the given amount."""
        super().labels().inc(*args)


class SummaryWrapper(CustomMetricsWrapperBase):
    def __init__(self, obj):
        super().__init__(obj)

    def observe(self, *args):
        """Observe the given amount.

        The amount is usually positive or zero. Negative values are
        accepted but prevent current versions of Prometheus from
        properly detecting counter resets in the sum of
        observations. See
        https://prometheus.io/docs/practices/histograms/#count-and-sum-of-observations
        for details.
        """
        super().labels().observe(*args)

    def time(self):
        """Time a block of code or function, and observe the duration in seconds.

        Can be used as a function decorator or context manager.
        """
        return super().labels().time()


class HistogramWrapper(CustomMetricsWrapperBase):
    def __init__(self, obj):
        super().__init__(obj)

    def observe(self, *args):
        """Observe the given amount.

        The amount is usually positive or zero. Negative values are
        accepted but prevent current versions of Prometheus from
        properly detecting counter resets in the sum of
        observations. See
        https://prometheus.io/docs/practices/histograms/#count-and-sum-of-observations
        for details.
        """
        super().labels().observe(*args)

    def time(self):
        """Time a block of code or function, and observe the duration in seconds.

        Can be used as a function decorator or context manager.
        """
        return super().labels().time()


class GaugeWrapper(CustomMetricsWrapperBase):
    def __init__(self, obj):
        super().__init__(obj)

    def inc(self, *args):
        """Increment gauge by the given amount."""
        super().labels().inc(*args)

    def dec(self, *args):
        """Decrement gauge by the given amount."""
        super().labels().dec(*args)

    def set(self, *args):
        """Set gauge to the current unixtime."""
        super().labels().set(*args)

    def track_inprogress():
        """Track inprogress blocks of code or functions.

        Can be used as a function decorator or context manager.
        Increments the gauge when the code is entered,
        and decrements when it is exited.
        """
        super().labels().track_inprogress()




import os

from urllib.parse import urlparse

from opentelemetry import trace, metrics
from opentelemetry.sdk.resources import Resource

from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import SimpleSpanProcessor, BatchSpanProcessor, ConsoleSpanExporter

from opentelemetry.exporter.otlp.proto.grpc.metric_exporter import OTLPMetricExporter
from opentelemetry.sdk.metrics import MeterProvider
from opentelemetry.sdk.metrics.export import PeriodicExportingMetricReader, ConsoleMetricExporter

from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import OTLPSpanExporter

from opentelemetry.instrumentation.requests import RequestsInstrumentor
from opentelemetry.instrumentation.flask import FlaskInstrumentor
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from opentelemetry.instrumentation.pymongo import PymongoInstrumentor
from opentelemetry.instrumentation.kafka import KafkaInstrumentor
from opentelemetry.instrumentation.pika import PikaInstrumentor

from opentelemetry.semconv.trace import SpanAttributes

from .metrics.metrics import PrometheusMiddleware
from .metrics.collector import MetricCollector


DEBUG_LOG_OTEL_TO_PROVIDER = False
DISABLE_TRACE = (os.getenv("OTEL_DISABLE_TRACE", False) == True)
DEFAULT_EXCLUDED_ROUTES = "metric,healthy,healthz,readyz,livez,startz,favicon.ico"
ENV_EXCLUDED_URLS = os.getenv("OTEL_PYTHON_EXCLUDED_URLS", "")


class TNObserver:

    metric_client = MetricCollector.histogram(
        "http_client_duration_milliseconds_count",
        "metric for measurement to other api/service",
        labels=["http_request_method", "service_address", "http_response_status_code"]
    )

    def __init__(self, service_name: str, service_namespace:str=None, service_version=None):
        service_namespace = service_namespace if service_namespace else "local" 
        self.resource = TNObserver.setup_resource(service_name, service_namespace, service_version)


    @classmethod
    def with_default_service_info(cls):
        service_name = os.getenv("SERVICE_NAME", "")
        service_namespace = os.getenv("SERVICE_NAME_PREFIX", "local")
        service_version = os.getenv("APP_VERSION")
        return cls(service_name, service_namespace, service_version)
        

    @staticmethod
    def get_service_name():
        prefix = os.getenv("SERVICE_NAME_PREFIX", None)
        name = os.getenv("SERVICE_NAME", "")

        if prefix:
            return f"{prefix}/{name}"
        return name


    @property
    def resources(self,) -> Resource:
        return self.resource


    @staticmethod
    def setup_resource(service_name: str, service_namespace: str, service_version: str):
        resource =  Resource.create({
            "service.name": f"{service_namespace}/{service_name}",
            "service.version": service_version if service_version else "1.0.0",
            "service.namespace": service_namespace,
            "app.name": service_name,
        })
        return resource
    


    @staticmethod
    def setup_trace(name, resource=None):

        trace_provider = TracerProvider(resource=resource)
        trace.set_tracer_provider(trace_provider)

        if not DISABLE_TRACE:
            if DEBUG_LOG_OTEL_TO_PROVIDER :
                span_process = SimpleSpanProcessor(ConsoleSpanExporter())
            else:
                otel_endpoint_url = os.getenv("OTEL_EXPORTER_OTLP_ENDPOINT", "http://localhost:4317")
                # otlp_span_exporter = OTLPSpanExporter(endpoint=otel_endpoint_url,headers=otel_http_headers)
                otlp_span_exporter = OTLPSpanExporter(endpoint=otel_endpoint_url)
                span_process = BatchSpanProcessor(otlp_span_exporter)
                # trace_provider.add_span_processor(BatchSpanProcessor(otlp_span_exporter))

            trace.get_tracer_provider().add_span_processor(span_process)
            tracer = trace.get_tracer(name)

            return tracer
        
        return None


    @staticmethod
    def setup_metrics(name, resource=None):
        # print (DEBUG_LOG_OTEL_TO_PROVIDER)
        if DEBUG_LOG_OTEL_TO_PROVIDER :
            metric_provider = MeterProvider(
                metric_readers=[PeriodicExportingMetricReader(ConsoleMetricExporter(), export_interval_millis=5000)],
                resource=resource
                )
        else:
            otel_endpoint_url = os.getenv('OTEL_EXPORTER_OTLP_ENDPOINT', "http://localhost:4317")
            metric_provider = MeterProvider(
                metric_readers=[PeriodicExportingMetricReader(OTLPMetricExporter(endpoint=otel_endpoint_url), export_interval_millis=50000)],
                resource=resource
                )

        metrics.set_meter_provider(metric_provider)
        meter = metrics.get_meter(name)
        return meter
    

    @staticmethod
    def _str_exclude_urls(excluded_urls):

        str_excludes_routes = DEFAULT_EXCLUDED_ROUTES
        if ENV_EXCLUDED_URLS != "":
            str_excludes_routes += f",{ENV_EXCLUDED_URLS}"
            
        if isinstance(excluded_urls, list):
            temp_urls = []
            for url in excluded_urls:
                if isinstance(url, dict) and "route" in url:
                    url = url["route"][1:] if url["route"][0] == "/" else url["route"]
                    temp_urls.append(url)
                elif isinstance(url, str):
                    url = url [1:] if url[0] == "/" else url
                    temp_urls.append(url)

            str_urls = ",".join(temp_urls)
            return str_excludes_routes + "," + str_urls
        elif isinstance(excluded_urls, str):
            return str_excludes_routes + "," + excluded_urls
        else:
            return str_excludes_routes


    @staticmethod
    def flask_instrumentation(app, excluded_urls=None):
        excluded_urls = TNObserver._str_exclude_urls(excluded_urls)
        return FlaskInstrumentor().instrument_app(app, excluded_urls=excluded_urls)
    
    
    @staticmethod
    def fast_instrumentation(app, excluded_urls=None):
        excluded_urls = TNObserver._str_exclude_urls(excluded_urls)
        return FastAPIInstrumentor().instrument_app(app, excluded_urls=excluded_urls)
    
    
    @staticmethod
    def pymongo_instrumentation():
        return PymongoInstrumentor().instrument()
    

    @staticmethod
    def _requests_request_hook(span: trace.Span, req):
        url = req.url
        o = urlparse(url)
        if span and span.is_recording():
            span.update_name(f"{req.method} {o.path}")
            span.set_attribute(SpanAttributes.HTTP_ROUTE, o.path)


    @staticmethod
    def _requests_response_hook(span: trace.Span, req, res):
        duration = int(res.elapsed.total_seconds() * 1000)
        TNObserver.metric_client.labels(req.method ,req.url, 
                                        res.status_code, 
                                        ).observe(duration)
    

    @staticmethod
    def requests_instrumentation():
        return RequestsInstrumentor().instrument(
            request_hook=TNObserver._requests_request_hook, 
            response_hook=TNObserver._requests_response_hook)
    

    @staticmethod
    def kafka_instrumentation():
        return KafkaInstrumentor().instrument()
    

    @staticmethod
    def pika_instrumentation():
        return PikaInstrumentor().instrument()
    

    @staticmethod
    def pika_instrumentor():
        return PikaInstrumentor()
    

    @staticmethod
    def _list_str_exclude_urls(str_exclude):
        return [{"route": f"/{url}"} for url in str_exclude.split(",")]
    

    @staticmethod
    def _list_exclude_urls(excluded_urls):
        list_excluded_route = []

        list_excluded_route.extend(TNObserver._list_str_exclude_urls(DEFAULT_EXCLUDED_ROUTES))
        list_excluded_route.extend(TNObserver._list_str_exclude_urls(ENV_EXCLUDED_URLS))

        if isinstance(excluded_urls, str):
            list_excluded_route.extend(TNObserver._list_str_exclude_urls(excluded_urls))

        if isinstance(excluded_urls, list):
            for url in excluded_urls:
                if isinstance(url, str):
                    list_excluded_route.append({"route": f"/{url}"})
                elif isinstance(url, dict):
                    url['route'] = f"/{url['route']}" if url['route'][0] != "/" else url['route']
                    list_excluded_route.append(url)
                    
        return list_excluded_route
    

    @staticmethod
    def register_metrics(app, resource: Resource, is_multiprocessing: bool=False, excluded_urls:list=[]):
        namespace = resource.attributes['service.namespace']
        service_name = resource.attributes['app.name']
        prom_middleware = PrometheusMiddleware(app, is_multiprocessing=is_multiprocessing, 
                                               service_name=service_name, namespace=namespace)
        
        list_exclude = TNObserver._list_exclude_urls(excluded_urls)
        prom_middleware.add_exclude(list_exclude)
        prom_middleware.register_metrics()



from random import random, randint
import unittest
import os
from pathlib import Path


import time
from flask import Flask
from flask_cors import CORS
from fastapi import FastAPI
from fastapi.testclient import TestClient


from src.thinknet_observer import TNObserver
from src.thinknet_observer import SingletonMeta
from src.thinknet_observer import MetricCollector

from src.thinknet_observer.metrics.config import NAMESPACE, SERVICENAME


CUSTOM_GAUGE = MetricCollector.gauge(
    "CUSTOM_GAUGE", "desc of CUSTOM_GAUGE", ["something"]
)
CUSTOM_GAUGE_NOLABEL = MetricCollector.gauge(
    "CUSTOM_GAUGE_NOLABEL", "desc of CUSTOM_GAUGE_NOLABEL"
)
CUSTOM_SUMMARY = MetricCollector.summary(
    "CUSTOM_SUMMARY", "desc of CUSTOM_SUMMARY", ["something"]
)
CUSTOM_SUMMARY_NOLABEL = MetricCollector.summary(
    "CUSTOM_SUMMARY_NOLABEL",
    "desc of CUSTOM_SUMMARY_NOLABEL",
)
CUSTOM_HISTOGRAM = MetricCollector.histogram(
    "CUSTOM_HISTOGRAM", "desc of CUSTOM_HISTOGRAM", ["something"]
)
CUSTOM_HISTOGRAM_NOLABEL = MetricCollector.histogram(
    "CUSTOM_HISTOGRAM_NOLABEL", "desc of CUSTOM_HISTOGRAM_NOLABEL"
)
CUSTOM_HISTOGRAM_NOLABEL_CUSTOMBUCKET = MetricCollector.histogram(
    "CUSTOM_HISTOGRAM_NOLABEL_CUSTOMBUCKET",
    "desc of CUSTOM_HISTOGRAM_NOLABEL_CUSTOMBUCKET",
    buckets=[0.5, 0.75, 1],
)
CUSTOM_COUNTER = MetricCollector.counter(
    "CUSTOM_COUNTER", "desc of CUSTOM_COUNTER", ["something"]
)
CUSTOM_COUNTER_NOLABEL = MetricCollector.counter(
    "CUSTOM_COUNTER_NOLABEL", "desc of CUSTOM_COUNTER_NOLABEL"
)


class PrometheusMiddlewareFastapiTestCase(unittest.TestCase):

    app_fastapi = FastAPI()

    @app_fastapi.get("/healthz")
    def health_check():
        return {"msg": "ok healthz"}

    @app_fastapi.get("/healthy")
    def healthy_check():
        return {"msg": "ok healthy"}

    @app_fastapi.get("/sleep")
    def get_sleep():
        time.sleep(1)
        return {"msg": "sleep for 1 sec"}

    @app_fastapi.post("/banana/{number}")
    def banana_number(number):
        return {"msg": f"add {number} banana"}

    @app_fastapi.post("/banana/{number}/eat")
    def eat_banana(number):
        return {"msg": f"ate {number} banana"}

    @app_fastapi.post("/inc_gauge/{number}")
    def inc_gauge(number):
        CUSTOM_GAUGE.labels("something'value").inc(float(number))
        CUSTOM_GAUGE_NOLABEL.inc(float(number))
        return {"msg": f"inc {number}"}

    @app_fastapi.post("/dec_gauge/{number}")
    def dec_gauge(number):
        CUSTOM_GAUGE.labels("something'value").dec(float(number))
        CUSTOM_GAUGE_NOLABEL.dec(float(number))
        return {"msg": f"dec {number}"}

    @app_fastapi.post("/summary_observe/{number}")
    def summary_observe(number):
        CUSTOM_SUMMARY.labels("something'value").observe(float(number))
        CUSTOM_SUMMARY_NOLABEL.observe(float(number))
        return {"msg": f"summary_observe {number}"}

    @app_fastapi.post("/histogram_observe/{number}")
    def histogram_observe(number):
        CUSTOM_HISTOGRAM.labels("something'value").observe(float(number))
        CUSTOM_HISTOGRAM_NOLABEL.observe(float(number))
        return {"msg": f"histogram_observe {number}"}

    @app_fastapi.post("/histogram_observe2/{number}")
    def histogram_observe2(number):
        CUSTOM_HISTOGRAM_NOLABEL_CUSTOMBUCKET.observe(float(number))
        return {"msg": f"histogram_observe {number}"}

    @app_fastapi.post("/count/{number}")
    def count_metric(number):
        CUSTOM_COUNTER.labels("something'value").inc(float(number))
        CUSTOM_COUNTER_NOLABEL.inc(float(number))
        return {"msg": f"count {number}"}

    def new_instance(self):
        SingletonMeta._instances = {}

    def clear_MULTIPROC_DIR(self):

        MULTIPROC_DIR = os.environ["PROMETHEUS_MULTIPROC_DIR"]
        if MULTIPROC_DIR:
            file_path = Path(MULTIPROC_DIR)
            if os.path.isdir(file_path):
                for f in os.listdir(file_path):
                    os.remove(os.path.join(file_path, f))
            else:
                os.mkdir(file_path)

    # NOTE: setUp method are called everytime before test cases.
    def setUp(self):
        self.app = self.app_fastapi
        TNObserver.register_metrics(self.app, is_multiprocessing=True)
        # self.middleware = PrometheusMiddleware(
        #     self.app_fastapi, is_multiprocessing=True
        # )
        # self.middleware.register_metrics()
        # self.middleware.clear_default_metrics()
        # self.middleware.exclude_list = self.middleware.default_exclude_list

    def tearDown(self):
        self.app = None
        self.middleware = None

    def test_client(self):
        test_client = TestClient(self.app)
        return test_client

    def test_case_001_exclude_list(self):
        """test get default exclude"""
        if (
            not isinstance(self.middleware.app, FastAPI)
            and SingletonMeta._instances is not {}
        ):
            self.new_instance()
        result = self.middleware.exclude_list
        self.assertEqual(result, [{"route": "/metrics/"}, {"route": "/metrics"}])

    def test_case_002_add_exclude(self):
        """add simple valid exclude to the middleware"""

        list_exclude = [{"route": "/first"}, {"route": "/second"}]
        self.middleware.add_exclude(list_exclude)
        result = self.middleware.exclude_list
        self.assertEqual(
            result,
            [
                {"route": "/metrics/"},
                {"route": "/metrics"},
                {"route": "/first"},
                {"route": "/second"},
            ],
        )

    def test_case_003_add_exclude(self):
        """add complex valid exclude to the middleware"""

        list_exclude = [
            {"route": "/third", "method": "get", "status_code": "200"},
            {"route": "/forth", "status_code": "400"},
        ]
        self.middleware.add_exclude(list_exclude)
        result = self.middleware.exclude_list
        self.assertEqual(
            result,
            [
                {"route": "/metrics/"},
                {"route": "/metrics"},
                {"route": "/third", "method": "get", "status_code": "200"},
                {"route": "/forth", "status_code": "400"},
            ],
        )

    def test_case_004_add_exclude(self):
        """add simple invalid exclude to the middleware"""

        list_exclude = [{"path": "/first"}, {"path": "/second"}]

        with self.assertRaises(ValueError):
            self.middleware.add_exclude(list_exclude)

    def test_case_005_add_exclude(self):
        """add simple invalid exclude to the middleware"""

        list_exclude = [{"path": "/first"}, {"path": "/second", "method": "post"}]
        with self.assertRaises(ValueError):
            self.middleware.add_exclude(list_exclude)

    def test_case_006_is_exclude(self):
        """test whether given param is exclude or not by give non-exclude param"""

        result = self.middleware.exclude_list
        self.assertEqual(result, [{"route": "/metrics/"}, {"route": "/metrics"}])
        self.assertFalse(self.middleware.is_exclude("get", "/first", "200"))

    def test_case_007_is_exclude(self):
        """test whether given param is exclude or not by give default-exclude param"""

        result = self.middleware.exclude_list
        self.assertEqual(result, [{"route": "/metrics/"}, {"route": "/metrics"}])
        self.assertTrue(self.middleware.is_exclude("get", "/metrics", "200"))

    def test_case_008_is_exclude(self):
        """test whether given param is exclude or not by give custom-exclude param"""

        result = self.middleware.exclude_list
        self.assertEqual(result, [{"route": "/metrics/"}, {"route": "/metrics"}])

        list_exclude = [
            {"route": "/first", "method": "get", "status_code": "200"},
            {"route": "/second", "status_code": "400"},
            {"status_code": "500"},
        ]
        self.middleware.add_exclude(list_exclude)
        self.assertTrue(self.middleware.is_exclude("get", "/first", "200"))
        self.assertFalse(self.middleware.is_exclude("get", "/first", "400"))
        self.assertTrue(self.middleware.is_exclude("get", "/second", "400"))
        self.assertTrue(self.middleware.is_exclude("get", "/second", "400"))
        self.assertTrue(self.middleware.is_exclude("any", "/whatever", "500"))

    def test_case_009_call_healthz(self):
        """call route /healthz and check response"""

        self.middleware.add_exclude([{"route": "/healthz"}])
        response = self.test_client().get("/healthz")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == '{"msg":"ok healthz"}')

        list_unexpected_data = [
            'http_request_duration_ht_ms_count{http_status="200",method="get",route="/healthz"} 1.0',
            'http_request_duration_summary_ms_count{http_status="200",method="get",route="/healthz"} 1.0',
            'http_request_total{http_status="200",method="get",route="/healthz"} 1.0',
        ]
        response = self.test_client().get("/metrics")
        metrics_data = str(response.text)
        self.assertTrue(
            not any(
                expected_data in metrics_data for expected_data in list_unexpected_data
            )
        )

    def test_case_010_call_healthy(self):
        """call route /healthy and check response"""

        response = self.test_client().get("/healthy")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == '{"msg":"ok healthy"}')

    def test_case_011_call_metrics(self):
        """call route /healthz and then check /metrics responses"""
        # 1st call
        response = self.test_client().get("/healthz")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == '{"msg":"ok healthz"}')

        list_expected_data_1 = [
            'http_request_duration_ms_ht_count{{http_status="200",method="get",namespace="{NAMESPACE}",route="/healthz",serviceName="{SERVICENAME}"}} 1.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
            'http_request_duration_ms_summary_count{{http_status="200",method="get",namespace="{NAMESPACE}",route="/healthz",serviceName="{SERVICENAME}"}} 1.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
            'http_request_total{{http_status="200",method="get",namespace="{NAMESPACE}",route="/healthz",serviceName="{SERVICENAME}"}} 1.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)

        self.assertTrue(response.status_code == 200)
        self.assertTrue(
            all(expected_data in metrics_data for expected_data in list_expected_data_1)
        )

        # 2nd call
        response = self.test_client().get("/healthz")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == '{"msg":"ok healthz"}')
        list_expected_data_2 = [
            'http_request_duration_ms_ht_count{{http_status="200",method="get",namespace="{NAMESPACE}",route="/healthz",serviceName="{SERVICENAME}"}} 2.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
            'http_request_duration_ms_summary_count{{http_status="200",method="get",namespace="{NAMESPACE}",route="/healthz",serviceName="{SERVICENAME}"}} 2.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
            'http_request_total{{http_status="200",method="get",namespace="{NAMESPACE}",route="/healthz",serviceName="{SERVICENAME}"}} 2.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)

        self.assertTrue(response.status_code == 200)
        self.assertTrue(
            all(expected_data in metrics_data for expected_data in list_expected_data_2)
        )

    def test_case_012_call_metrics(self):
        """call route /sleep and then check /metrics responses"""

        response = self.test_client().get("/sleep")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == '{"msg":"sleep for 1 sec"}')

        list_expected_data_1 = [
            'http_request_duration_ms_ht_count{{http_status="200",method="get",namespace="{NAMESPACE}",route="/sleep",serviceName="{SERVICENAME}"}} 1.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
            'http_request_duration_ms_summary_count{{http_status="200",method="get",namespace="{NAMESPACE}",route="/sleep",serviceName="{SERVICENAME}"}} 1.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
            'http_request_total{{http_status="200",method="get",namespace="{NAMESPACE}",route="/sleep",serviceName="{SERVICENAME}"}} 1.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)
        timed_sleep = self.middleware.PROMETHEUS_REGISTRY.get_sample_value(
            "http_request_duration_ms_ht_sum",
            labels={
                "http_status": "200",
                "method": "get",
                "namespace": f"{NAMESPACE}",
                "route": "/sleep",
                "serviceName": f"{SERVICENAME}",
            },
        )

        self.assertTrue(response.status_code == 200)
        self.assertTrue(
            all(expected_data in metrics_data for expected_data in list_expected_data_1)
        )
        self.assertTrue(1 <= timed_sleep <= 2)

    def test_case_013_call_metrics(self):
        """call dynamics routes and then check /metrics responses"""
        # 1st call
        num = randint(1, 10)
        response = self.test_client().post(f"/banana/{num}")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == f'{{"msg":"add {num} banana"}}')

        list_expected_data_1 = [
            'http_request_duration_ms_ht_count{{http_status="200",method="post",namespace="{NAMESPACE}",route="/banana/{{number}}",serviceName="{SERVICENAME}"}} 1.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
            'http_request_duration_ms_summary_count{{http_status="200",method="post",namespace="{NAMESPACE}",route="/banana/{{number}}",serviceName="{SERVICENAME}"}} 1.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
            'http_request_total{{http_status="200",method="post",namespace="{NAMESPACE}",route="/banana/{{number}}",serviceName="{SERVICENAME}"}} 1.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)
        self.assertTrue(response.status_code == 200)
        self.assertTrue(
            all(expected_data in metrics_data for expected_data in list_expected_data_1)
        )

        # 2nd call
        num = randint(1, 10)
        response = self.test_client().post(f"/banana/{num}/eat")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == f'{{"msg":"ate {num} banana"}}')
        list_expected_data_1 = [
            'http_request_duration_ms_ht_count{{http_status="200",method="post",namespace="{NAMESPACE}",route="/banana/{{number}}/eat",serviceName="{SERVICENAME}"}} 1.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
            'http_request_duration_ms_summary_count{{http_status="200",method="post",namespace="{NAMESPACE}",route="/banana/{{number}}/eat",serviceName="{SERVICENAME}"}} 1.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
            'http_request_total{{http_status="200",method="post",namespace="{NAMESPACE}",route="/banana/{{number}}/eat",serviceName="{SERVICENAME}"}} 1.0'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)
        self.assertTrue(response.status_code == 200)
        self.assertTrue(
            all(expected_data in metrics_data for expected_data in list_expected_data_1)
        )

    def test_case_014_call_custom_gauge_metrics(self):
        """call routes and then check /metrics responses for CUSTOM_GAUGE and CUSTOM_GAUGE_NOLABEL"""

        # 1st call
        num1 = randint(1, 10)
        response = self.test_client().post(f"/inc_gauge/{num1}")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == f'{{"msg":"inc {num1}"}}')
        list_expected_data = [
            'CUSTOM_GAUGE{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}",something="something\'value"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(num1)
            ),
            'CUSTOM_GAUGE_NOLABEL{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(num1)
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)
        self.assertTrue(response.status_code == 200)
        # self.assertTrue(
        #     all(expected_data in metrics_data for expected_data in list_expected_data)
        # )

        # 2nd call
        num2 = randint(1, 10)
        response = self.test_client().post(f"/dec_gauge/{num2}")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == f'{{"msg":"dec {num2}"}}')
        list_expected_data = [
            'CUSTOM_GAUGE{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}",something="something\'value"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(num1 - num2)
            ),
            'CUSTOM_GAUGE_NOLABEL{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(num1 - num2)
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)
        self.assertTrue(response.status_code == 200)
        # self.assertTrue(
        #     all(expected_data in metrics_data for expected_data in list_expected_data)
        # )

        CUSTOM_GAUGE.clear()
        CUSTOM_GAUGE_NOLABEL.clear()

    def test_case_015_call_custom_summary_metrics(self):
        """call routes and then check /metrics responses for CUSTOM_SUMMARY_NOLABEL and CUSTOM_SUMMARY_NOLABEL"""

        # 1st call
        num1 = random()
        response = self.test_client().post(f"/summary_observe/{num1}")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == f'{{"msg":"summary_observe {num1}"}}')
        list_expected_data = [
            'CUSTOM_SUMMARY_sum{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}",something="something\'value"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(num1)
            ),
            'CUSTOM_SUMMARY_NOLABEL_sum{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(num1)
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)
        self.assertTrue(response.status_code == 200)
        self.assertTrue(
            all(expected_data in metrics_data for expected_data in list_expected_data)
        )

        # 2nd call
        num2 = random()
        response = self.test_client().post(f"/summary_observe/{num2}")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == f'{{"msg":"summary_observe {num2}"}}')
        list_expected_data = [
            'CUSTOM_SUMMARY_sum{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}",something="something\'value"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(num1 + num2)
            ),
            'CUSTOM_SUMMARY_NOLABEL_sum{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(num1 + num2)
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)
        self.assertTrue(response.status_code == 200)
        self.assertTrue(
            all(expected_data in metrics_data for expected_data in list_expected_data)
        )

        CUSTOM_SUMMARY.clear()
        CUSTOM_SUMMARY_NOLABEL.clear()

    def test_case_016_call_custom_histogram_metrics(self):
        """call routes and then check /metrics responses for CUSTOM_HISTOGRAM and CUSTOM_HISTOGRAM_NOLABEL"""

        # 1st call
        num1 = 0.06
        response = self.test_client().post(f"/histogram_observe/{num1}")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == f'{{"msg":"histogram_observe {num1}"}}')
        list_expected_data = [
            'CUSTOM_HISTOGRAM_bucket{{le="0.05",namespace="{NAMESPACE}",serviceName="{SERVICENAME}",something="something\'value"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(0)
            ),
            'CUSTOM_HISTOGRAM_bucket{{le="0.1",namespace="{NAMESPACE}",serviceName="{SERVICENAME}",something="something\'value"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(1)
            ),
            'CUSTOM_HISTOGRAM_NOLABEL_bucket{{le="0.05",namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(0)
            ),
            'CUSTOM_HISTOGRAM_NOLABEL_bucket{{le="0.1",namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(1)
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)
        self.assertTrue(response.status_code == 200)
        self.assertTrue(
            all(expected_data in metrics_data for expected_data in list_expected_data)
        )

        # # 2nd call
        num2 = 0.12
        response = self.test_client().post(f"/histogram_observe/{num2}")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == f'{{"msg":"histogram_observe {num2}"}}')
        list_expected_data = [
            'CUSTOM_HISTOGRAM_bucket{{le="0.05",namespace="{NAMESPACE}",serviceName="{SERVICENAME}",something="something\'value"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(0)
            ),
            'CUSTOM_HISTOGRAM_bucket{{le="0.1",namespace="{NAMESPACE}",serviceName="{SERVICENAME}",something="something\'value"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(1)
            ),
            'CUSTOM_HISTOGRAM_bucket{{le="0.5",namespace="{NAMESPACE}",serviceName="{SERVICENAME}",something="something\'value"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(2)
            ),
            'CUSTOM_HISTOGRAM_NOLABEL_bucket{{le="0.05",namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(0)
            ),
            'CUSTOM_HISTOGRAM_NOLABEL_bucket{{le="0.1",namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(1)
            ),
            'CUSTOM_HISTOGRAM_NOLABEL_bucket{{le="0.5",namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(2)
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)
        self.assertTrue(response.status_code == 200)
        self.assertTrue(
            all(expected_data in metrics_data for expected_data in list_expected_data)
        )
        CUSTOM_HISTOGRAM.clear()
        CUSTOM_HISTOGRAM_NOLABEL.clear()

    def test_case_017_call_custom_counter_metrics(self):
        # 1st call
        num1 = randint(1, 10)
        response = self.test_client().post(f"/count/{num1}")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == f'{{"msg":"count {num1}"}}')
        list_expected_data = [
            'CUSTOM_COUNTER_total{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}",something="something\'value"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=num1
            ),
            'CUSTOM_COUNTER_total{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}",something="something\'value"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=num1
            ),
            'CUSTOM_COUNTER_NOLABEL_total{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=num1
            ),
            'CUSTOM_COUNTER_NOLABEL_total{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=num1
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)
        self.assertTrue(response.status_code == 200)
        self.assertTrue(
            all(expected_data in metrics_data for expected_data in list_expected_data)
        )

        # 2nd call
        num2 = randint(1, 10)
        response = self.test_client().post(f"/count/{num2}")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == f'{{"msg":"count {num2}"}}')
        list_expected_data = [
            'CUSTOM_COUNTER_total{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}",something="something\'value"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=num2 + num1
            ),
            'CUSTOM_COUNTER_total{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}",something="something\'value"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=num2 + num1
            ),
            'CUSTOM_COUNTER_NOLABEL_total{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=num2 + num1
            ),
            'CUSTOM_COUNTER_NOLABEL_total{{namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=num2 + num1
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)
        self.assertTrue(response.status_code == 200)
        self.assertTrue(
            all(expected_data in metrics_data for expected_data in list_expected_data)
        )
        CUSTOM_COUNTER.clear()
        CUSTOM_COUNTER_NOLABEL.clear()

    def test_case_018_call_custom_histogram_bucket_metrics(self):
        """call routes and then check /metrics responses for CUSTOM_HISTOGRAM_NOLABEL_CUSTOMBUCKET"""

        # 1st call
        num1 = 0.06
        response = self.test_client().post(f"/histogram_observe2/{num1}")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == f'{{"msg":"histogram_observe {num1}"}}')
        list_expected_data = [
            'CUSTOM_HISTOGRAM_NOLABEL_CUSTOMBUCKET_bucket{{le="0.5",namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(1)
            ),
            'CUSTOM_HISTOGRAM_NOLABEL_CUSTOMBUCKET_bucket{{le="0.75",namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(1)
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)
        self.assertTrue(response.status_code == 200)
        self.assertTrue(
            all(expected_data in metrics_data for expected_data in list_expected_data)
        )

        # # 2nd call
        num2 = 0.8
        response = self.test_client().post(f"/histogram_observe2/{num2}")
        self.assertTrue(response.status_code == 200)
        self.assertTrue(response.text == f'{{"msg":"histogram_observe {num2}"}}')
        list_expected_data = [
            'CUSTOM_HISTOGRAM_NOLABEL_CUSTOMBUCKET_bucket{{le="0.5",namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(1)
            ),
            'CUSTOM_HISTOGRAM_NOLABEL_CUSTOMBUCKET_bucket{{le="0.75",namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(1)
            ),
            'CUSTOM_HISTOGRAM_NOLABEL_CUSTOMBUCKET_bucket{{le="1.0",namespace="{NAMESPACE}",serviceName="{SERVICENAME}"}} {number}'.format(
                NAMESPACE=NAMESPACE, SERVICENAME=SERVICENAME, number=float(2)
            ),
        ]
        response_metrics = self.test_client().get("/metrics")
        metrics_data = str(response_metrics.text)
        self.assertTrue(response.status_code == 200)
        self.assertTrue(
            all(expected_data in metrics_data for expected_data in list_expected_data)
        )
        CUSTOM_HISTOGRAM_NOLABEL_CUSTOMBUCKET.clear()
        self.clear_MULTIPROC_DIR()


if __name__ == "__main__":
    unittest.main()



import logging
import time
import re

from fastapi import Request, HTTPException
from starlette.responses import Response
from starlette.types import ASGIApp
from starlette.middleware.base import BaseHTTPMiddleware

from ..loggers.loggers import TNAccessLogger, TNLogger, UnExpectedError
from ..metrics.metrics import PrometheusMiddleware
from ..utils.http import HttpContext

from . import get_resource_info


class FastObserverMiddleware(BaseHTTPMiddleware):
# class FastObserverMiddleware():

    def __init__(self, app: ASGIApp, resource, is_multiprocessing: bool=False, excluded_urls:list=[] ) -> None:
        super().__init__(app)
        self._resource = resource
        self._is_multiprocessing = is_multiprocessing
        self._excluded_urls = excluded_urls
        
        self._access_log = TNAccessLogger(self._resource)
        # self._setup_metrics(app)


    def _setup_metrics (self, app,):
        namespace, service_name = get_resource_info(self._resource)

        self._metrics = PrometheusMiddleware(app, namespace, service_name, 
                                            is_multiprocessing=self._is_multiprocessing)
        self._metrics.add_exclude(self._excluded_urls)
        self._metrics.register_fastapi_metrics()
    

    # async def __call__(self, request: Request, call_next) -> Response: 
    async def dispatch(self, request: Request, call_next) -> Response: 
        tick = time.time()
        response = await call_next(request)
        duration = time.time() - tick

        http_context = HttpContext(request.method, request.scope.get("path", request.url.path), request.headers.get('referer', ""),
                                request.client.host, duration, request.headers.get('user-agent',""),
                                request.app.routes, response.status_code )
        
        self._access_log.logging(http_context)
        # self._metrics.add_fastapi_metrics(http_context)
        
        return response
        
        


from starlette.requests import Request
from starlette.types import ASGIApp, Receive, Scope, Send


class MetricsRedirect:
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope.get("path"):
            if scope["path"] == "/metrics":
                scope["path"] = "/metrics/"
                scope["raw_path"] = b"/metrics/"
        await self.app(scope, receive, send)



# from .calulator import Calculator

from .metrics.collector import MetricCollector
from .metrics import config

# from .loggers.logger import logger
# from .loggers.accesslog import _create_logger, AccessLogAtoms
from .loggers.flask_logger import NotFoundError, FlaskErrorLogger, UnExpectedError
from .loggers.fast_logger import FastAPILogger
from .loggers.loggers import TNAPIError, TNLogger

from .middleware.flask_middleware import FlaskObserverMiddleware
from .middleware.fast_middleware import FastObserverMiddleware

from .utils import health
from .utils.http import HttpContext
from .otel import TNObserver

from .utils.singleton import SingletonMeta
from .utils.gunicorn_multiprocess import clear_multiproc_dir



import os
import time
import logging


# from werkzeug import Request, Response
from werkzeug.exceptions import InternalServerError, NotFound
from flask import jsonify

from .loggers import TNLogger, UnExpectedError, TNAPIError, NotFoundError


class FlaskErrorLogger:
    
    def __init__(self, app, resource=None) -> None:
        self.error_log = TNLogger("error", level=logging.ERROR, resource=resource)
        # self.init_app(app)


    def init_app(self, app):
        app.register_error_handler(404, NotFoundError)
        app.register_error_handler(500, UnExpectedError)


    def error_handler(self, exc: Exception):

        if isinstance(exc, TNAPIError):
            self._logs(exc)
            return jsonify(exc.errors), exc.http_status
        
        elif isinstance(exc, NotFound):
            obj_exc = NotFoundError()
            self._logs(obj_exc)
            return jsonify(obj_exc.errors), obj_exc.http_status
        else:
            # exc_type, exc_value, exc_tb = sys.exc_info() 
            # tb = traceback.TracebackException(exc_type, exc_value, exc_tb) 
            obj_exc = UnExpectedError("NO HANDLER ERROR", exc=exc)
            self._logs(obj_exc)

            return jsonify(obj_exc.errors), 500
        


    def _logs(self, obj_exc):

        if obj_exc.http_status == 500:
            self.error_log.error(obj_exc.msg_body, stack_info=True, exc_info=True, extra=obj_exc.attributes)
        else:
            self.error_log.error(obj_exc.msg_body, extra=obj_exc.attributes)




import json
import logging
import traceback
from logging import NOTSET

from time import strftime, gmtime
from typing import Union, Dict, Any

# from opentelemetry.instrumentation.logging import LoggingInstrumentor

# from opentelemetry._logs import set_logger_provider
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler, LogRecord
from opentelemetry.sdk._logs._internal import LogLimits
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor, SimpleLogRecordProcessor, ConsoleLogExporter
from opentelemetry.sdk.resources import Resource

from opentelemetry.trace.span import TraceFlags

from ..otel import TNObserver
from ..utils.http import HttpContext
"""
We have options in python for stdout (streamhandling) and file logging
File logging has options for a Rotating file based on size or time (daily)
or a watched file, which supports logrotate style rotation
Most of the changes happen in the handlers, lets define a few standards
"""


_TNLogLimits = LogLimits(
    max_attributes=LogLimits.UNSET,
    max_attribute_length=LogLimits.UNSET,
)


class TNLogRecord(LogRecord):
    
    def __init__(self, timestamp: int  = None, observed_timestamp: int  = None, 
                 trace_id: int  = None, span_id: int  = None, trace_flags: TraceFlags  = None, 
                 severity_text: str  = None, severity_number: int  = None, 
                 body: Any  = None, resource: Resource  = None, 
                 attributes: Any  = None, 
                 source: str  = None,
                 limits: LogLimits  = _TNLogLimits,
                 ):
        
        super().__init__(timestamp, observed_timestamp, trace_id, span_id, trace_flags, severity_text, severity_number, body, resource, attributes, limits)
        self.source = source
        self.severity_number = severity_number


    def to_json(self, indent=4) -> str:
        otel_log = json.loads(super().to_json(indent))
        otel_resource = otel_log['resource'].replace("'", "\"")
        otel_log['source'] = self.source
        otel_log['severity_number'] = self.severity_number

        # remove prefix '0x' from super class #
        otel_log['trace_id'] = otel_log['trace_id'][2:]
        otel_log['span_id'] = otel_log['span_id'][2:]

        otel_log['resource'] = json.loads(otel_resource)
        return json.dumps(otel_log)
    


class TNLoggingHandler(LoggingHandler):
    def __init__(self, level=logging.NOTSET, logger_provider=None, source='') -> None:
        super().__init__(level, logger_provider)
        self.source = source

    def _translate(self, record: logging.LogRecord) -> LogRecord:
        
        rc = super()._translate(record)
        severity_number = record.levelno

        return TNLogRecord(timestamp=rc.timestamp, observed_timestamp=rc.observed_timestamp, trace_id=rc.trace_id,
                           span_id=rc.span_id, trace_flags=rc.trace_flags, severity_text=rc.severity_text, 
                           severity_number=severity_number,
                           body=rc.body, resource=rc.resource, source=self.source, attributes=rc.attributes)
    
    

class TNLogger(logging.Logger):

    def __init__(self, name: str=None, level=logging.NOTSET, resource: Resource = None ):

        super().__init__(name, level)

        self.resource = resource
        logger_provider = LoggerProvider(resource=self.resource)

        exporter = ConsoleLogExporter()
        logger_provider.add_log_record_processor(SimpleLogRecordProcessor(exporter))

        # self.handler = LoggingHandler(logger_provider=logger_provider)
        self.handler = TNLoggingHandler(logger_provider=logger_provider, source=name)

        super().addHandler(self.handler)

    
    @classmethod
    def with_service_detail(cls, name, level):
        # resource = TNObserver.setup_resource(service_name, version)
        observer = TNObserver.with_default_service_info()
        return cls (name, level, resource=observer.resources)

    
    @staticmethod
    def get_timestamp():
        return strftime("%Y-%m-%dT%H:%M:%S",gmtime())

    


class TNAccessLogger:

    def __init__(self, resource, level=logging.INFO ) -> None:
        self.access_log = TNLogger("accesslog", level, resource=resource)


    def logging(self, http_context: HttpContext):
        attributes = {
            "remote_address": http_context.remote_address,
            "referrer": http_context.referrer if http_context.referrer else "",
            "http.request.method": http_context.method,
            "http.route": http_context.route,
            "http.response.status_code": http_context.status_code,
            "user_agent.original": f"{http_context.user_agent}",
            "duration": http_context.duration,
            "timestamp": TNLogger.get_timestamp()
        }
        self.access_log.info("access log", extra=attributes)




class TNAPIError(Exception):

    http_status: int
    message: str

    handler_code: str = None
    service_code: str = None
    external_service_code: str = None
    payload: Union[Dict [str, str], str, None] = None
    attributes: Union[Dict [str, str], str, None] = None
    stack: str = None
    resource: Resource = None

    error_log = TNLogger("error", logging.ERROR, resource)
    

    def __init__(self, handler_code: str, service_code: str, external_service_code: str, 
                 payload: Union[Dict [str, str], str, None], attributes: Union[Dict [str, str], str, None], stack: str):
        
        self.handler_code = handler_code
        self.service_code = service_code
        self.external_service_code = external_service_code
        self.payload = payload
        self.attributes = attributes
        self.stack = stack

        self.body = {
            "httpStatus": self.http_status,
            "handlerCode": self.handler_code if self.handler_code else "",
            "serviceCode": self.service_code if self.service_code else "",
            "externalServiceCode": self.external_service_code if self.external_service_code else "",
            "message": self.message,
            "payload": self.payload if self.payload else {},
            "stack": self.stack
        }
        
        # self._logs()

    
    @property
    def errors(self):
        return {"error": self.message}
    

    @property
    def msg_body(self):
        return self.body
    

    def _logs(self,):
        
        if self.http_status == 500:
            self.error_log.error(self.body, stack_info=True, exc_info=True, extra=self.attributes)
        else:
            self.error_log.error(self.body, extra=self.attributes)
        


class NotFoundError(TNAPIError):
    http_status = 404
    message = "The requested URL was not found on the server. If you entered the URL manually please check your spelling and try again"

    def __init__(self, handler_code: str=None, service_code: str=None, external_service_code: str=None, message: str=None, 
                 payload =None, attributes=None, stack:str=None):
        
        super().__init__(handler_code, service_code, external_service_code, payload, attributes, stack=stack)



class UnExpectedError(TNAPIError):
    http_status = 500
    message = "Internal Server Error"

    def __init__(self, handler_code: str=None, service_code: str=None, external_service_code: str=None, message: str=None, 
                 payload =None, attributes=None, exc: Exception=None):
        
        trace_stack = {
                "type": type(exc).__name__,
                "message": str(exc),
                "stacktrace": traceback.format_exc()
            }
        
        # super().message = message
        super().__init__(handler_code, service_code, external_service_code, payload, attributes, stack=trace_stack)





class HttpContext:
    def __init__(self, method: str, route: str, referrer: str, remote_address: str, 
                 duration: float, user_agent: str, url,status_code, ) -> None:
        self.method = method
        self.route = route
        self.referrer = referrer
        self.remote_address = remote_address
        self.duration = duration
        self.user_agent = user_agent
        self.url = url
        self.status_code = status_code


import time
import re, os

from flask import Flask
from flask import request as flask_request

from fastapi import FastAPI
from fastapi import Request as fastapi_request
from fastapi.middleware.wsgi import WSGIMiddleware

from prometheus_client import Counter, Histogram, Summary, Info
from prometheus_client import make_wsgi_app
from prometheus_client import REGISTRY as PROMETHEUS_REGISTRY
from prometheus_client import (
    CollectorRegistry,
    generate_latest,
    multiprocess,
    CONTENT_TYPE_LATEST,
)

from werkzeug.middleware.dispatcher import DispatcherMiddleware

from .fastapi_metrics_redirect import MetricsRedirect
from ..utils.singleton import SingletonMeta
from ..utils.bcolor import Bcolors
from ..utils.gunicorn_multiprocess import clear_multiproc_dir
from ..utils.http import HttpContext

from .config import (
    _DEFAULT_BUCKETS,
    _DEFAULT_LABELS,
    NAMESPACE,
    SERVICENAME,
    EXCLUDE_URLS,
    SECONDTOMS
)


class PrometheusMiddleware(metaclass=SingletonMeta):

    namespace = NAMESPACE
    service_name = SERVICENAME

    _DEFAULT_BUCKETS = _DEFAULT_BUCKETS

    METRICS_REQUEST_LATENCY_HISTOGRAM = Histogram(
        "http_server_duration_milliseconds",
        "Duration of HTTP requests in ms (Histogram)",
        labelnames=_DEFAULT_LABELS,
        buckets=_DEFAULT_BUCKETS,
    )

    METRICS_REQUEST_LATENCY_SUMMARY = Summary(
        "http_server_duration_milliseconds_summary",
        "Duration of HTTP requests in ms (Summary)",
        labelnames=_DEFAULT_LABELS,
    )

    METRICS_REQUEST_COUNT = Counter(
        "http_requests_total",
        "Number of HTTP requests",
        labelnames=_DEFAULT_LABELS,
    )

    METRICS_INFO = Info("app_version", "Application Version")

    def __init__(
        self,
        app,
        namespace: str = None,
        service_name: str = None,
        app_version: str = None,
        is_multiprocessing: bool = False,
    ):
        """
        :param app: web application framework can be either 'Flask' or 'fastapi' instance.
        :type app: 'astapi.applications.FastAPI' or 'flask.app.Flask'

        :param namespace: can be use to override namespace that read from .env
        :type namespace: str

        :param service_name: can be use to override service_name that read from .env
        :type service_name: str

        :param app_version: optional param can be use for track version of the application
        :type app_version: str
        """
        self.app = app
        self.exclude_list = [
            {"route": "/metrics/"},
            {"route": "/metrics"},
            {"route": "/metrics/{path}"},
        ]
        self._is_registered = False
        self.is_multiprocessing = is_multiprocessing
        if app_version:
            self.METRICS_INFO.info({"version": app_version})

        if namespace:
            self.namespace = namespace
        if service_name:
            self.service_name = service_name

        self._add_exclude_route_env()
        self._check_env(self.namespace, self.service_name)
        

    @staticmethod
    def _check_invalid_exclude_format(list_input):
        """
        raise an error when invalid exclude format is found.
        """
        tuple_keys = ("method", "route", "status_code")
        for idx, val in reversed(list(enumerate(list_input))):
            if not isinstance(val, dict):
                error_msg = f"{Bcolors.FAIL}ERROR : Invalid exclude found : '{val}' is not valid exclude format. exclude must be instance of dict(){Bcolors.ENDC}"
                list_input.pop(idx)
                raise TypeError(error_msg)
            elif not any(key in val.keys() for key in tuple_keys):
                error_msg = f'{Bcolors.FAIL}ERROR : Invalid exclude found : \'{val}\' is not valid exclude format. exclude must contain atleast 1 key from this list ["method","route","status_code"]{Bcolors.ENDC}'
                list_input.pop(idx)
                raise ValueError(error_msg)
            elif "route" in val.keys():
                if val["route"][0] != "/":
                    warning_msg = f"{Bcolors.WARNING}WARNING : exclude {val} route's value is not start with '/'. This exclude will likely to be ignored{Bcolors.ENDC}"
                    print(warning_msg)
        return list_input
    

    def _add_exclude_route_env(self,):
        str_exclude_url = EXCLUDE_URLS
        list_exclude_env = []

        if len(str_exclude_url) != 0:
            split_exclude_url = str_exclude_url.split(",")
            for url in split_exclude_url:
                list_exclude_env.append({'route': f"/{url}"})

        self.add_exclude(list_exclude_env)



    def _check_env(self, namespace, service_name):
        print (namespace)
        if namespace == "UNDEFINED" or namespace is None:
            print(
                f"{Bcolors.WARNING}WARNING: env variable : for label 'namespace' is not found.\nPlease provide SERVICE_NAME_PREFIX in .env file{Bcolors.ENDC}"
            )
        if service_name == "UNDEFINED" or service_name is None:
            print(
                f"{Bcolors.WARNING}WARNING: env variable : for label 'serviceName' is not found.\nPlease provide SERVICE_NAME in .env file{Bcolors.ENDC}"
            )

    def clear_default_metrics(self):
        """
        Clear all observed default metrics.
        Generally being use in testing only.
        """
        for default_metric in self.default_metrics:
            default_metric.clear()

    @property
    def app(self):
        return self._app

    @app.setter
    def app(self, new_app):
        if isinstance(new_app, Flask):
            self._app = new_app
            self.request = flask_request
        elif isinstance(new_app, FastAPI):
            self._app = new_app
            self.request = fastapi_request
        else:
            raise TypeError("app type have to be etiher 'Flask' or 'fastapi'.")

    @property
    def request(self):
        return self._request

    @request.setter
    def request(self, request_class):
        self._request = request_class

    @property
    def exclude_list(self):
        return self._exclude_list

    @exclude_list.setter
    def exclude_list(self, value: list):
        if isinstance(value, list):
            value = self._check_invalid_exclude_format(value)
            self._exclude_list = value
        else:
            raise TypeError("Invalid PrometheusMiddleware.exclude format")

    @property
    def default_metrics(self):
        """
        return a list of default metrics.
        """
        list_default_metrics = [
            self.METRICS_INFO,
            self.METRICS_REQUEST_LATENCY_HISTOGRAM,
            self.METRICS_REQUEST_LATENCY_SUMMARY,
            self.METRICS_REQUEST_COUNT,
        ]
        return list_default_metrics

    @property
    def default_exclude_list(self):
        return [{"route": "/metrics/"}, {"route": "/metrics"}]

    @property
    def PROMETHEUS_REGISTRY(self):
        """
        return REGISTRY object from prometheus client
        PROMETHEUS_REGISTRY can be use to access colleted metric
        Ex.
        PROMETHEUS_REGISTRY.get_sample_value(
        "http_request_duration_ht_ms_sum",
        labels={"http_status": "200", "method": "get", "route": "/sleep"},
        )
        this will retrieve data from "http_request_duration_ht_ms_sumq"
        """
        return PROMETHEUS_REGISTRY

    def add_exclude(self, added_exclude: list):
        """
        :param added_exclude: list of dict that contain atleast 1 key from this list ["method","route","status_code"]
        :type added_exclude: list(dict())

        Ex . [{"route":"/index"}] will exclude all metrics from route /index

        Ex . [{"route":"/index","status_code":"200"}] will exclude all metrics from route="/index" with status_code="200" (if error still collect metric)
        """
        list_buffer = self.exclude_list
        list_buffer.extend(added_exclude)
        self.exclude_list = list_buffer

    def is_exclude(self, method, route, status_code):
        """
        :param method: method of the endpoint
        :type method: str

        :param route: route of the endpoint
        :type route: str

        :param status_code: returned status of the request
        :type status_code: str
        """
        if not self.exclude_list:
            return False

        for exclude_item in self.exclude_list:
            cond_method = (
                True
                if exclude_item.get("method") is None
                else (exclude_item.get("method").lower() == method.lower())
            )
            cond_route = (
                True
                if exclude_item.get("route") is None
                else (exclude_item.get("route") == route)
            )
            cond_status_code = (
                True
                if exclude_item.get("status_code") is None
                else (exclude_item.get("status_code") == status_code)
            )

            if cond_method and cond_route and cond_status_code:
                return True

        return False
    
    
    @staticmethod
    def _app_multiprocess(environ, start_response):
        environ["PATH_INFO"] = "/metrics"
        registry = CollectorRegistry()
        multiprocess.MultiProcessCollector(registry)
        data = generate_latest(registry)

        status = "200 OK"
        response_headers = [
                        ("Content-type", CONTENT_TYPE_LATEST),
                        ("Content-Length", str(len(data))),
                    ]
        start_response(status, response_headers)
        return iter([data])
    

    @staticmethod
    def _make_wsgi_multiprocess_app():
        return PrometheusMiddleware._app_multiprocess


    def register_flask_metrics(self, ):
        self._is_registered = True

        if self.is_multiprocessing:
            clear_multiproc_dir()

            self.app.wsgi_app = DispatcherMiddleware(
                self.app.wsgi_app, {"/metrics": PrometheusMiddleware._make_wsgi_multiprocess_app()}
            )
        else:
            self.app.wsgi_app = self.app.wsgi_app = DispatcherMiddleware(
                self.app.wsgi_app, {"/metrics": make_wsgi_app()}
            )


    def add_metrics(self, method, url, status_code, duration):
        if not self.is_exclude(method, url, status_code):
            request_latency = int(duration) * SECONDTOMS

            self.METRICS_REQUEST_LATENCY_HISTOGRAM.labels(
                method,
                url,
                status_code,
                self.service_name,
                self.namespace,
            ).observe(request_latency)

            self.METRICS_REQUEST_LATENCY_SUMMARY.labels(
                method,
                url,
                status_code,
                self.service_name,
                self.namespace,
            ).observe(request_latency)

            self.METRICS_REQUEST_COUNT.labels(
                method,
                url,
                status_code,
                self.service_name,
                self.namespace,
            ).inc()
    
        
    def add_flask_metrics( self,  http_context: HttpContext):
        method = http_context.method
        url = http_context.url
        status_code = http_context.status_code
        duration = http_context.duration

        url_rule = (url.rule if url else http_context.route)

        self.add_metrics(method, url_rule, status_code, duration)


    def before_request(self):
        """
        Get start time of a request.
        This method get will call before a request processed.
        """
        self.request._tick= time.time()


    @staticmethod
    def get_path_format(app_routes, path):
        """loop through all endpoint to get parh_format"""
        path_format = path
        for route in app_routes:
            if re.search(route.path_regex, path) is not None:
                return route.path_format
        return path_format
        

    def register_fastapi_metrics(self, ):
        self._is_registered = True
        self.app.add_middleware(MetricsRedirect)

        if self.is_multiprocessing:
            clear_multiproc_dir()

            self.app.mount("/metrics", WSGIMiddleware(PrometheusMiddleware.make_wsgi_multiprocess_app()))
        else:
            self.app.mount("/metrics", WSGIMiddleware(make_wsgi_app()))

    
    def add_fastapi_metrics(self, request, response):
        url = self.get_path_format(request.app.routes, request.scope.get("path", request.url.path))
        raw_path = request.scope.get("raw_path", b"/").decode("utf-8")
        if raw_path == "/metrics/":
            request.scope["path"] = "/metrics"

        duration = int(time.time() - request._tick) * SECONDTOMS

        self.add_metrics(request.method, url, response.status_code, duration)
        return response

    
    def add_fastapi_middleware(self,):
        self.register_fastapi_metrics()

        @self.app.middleware("http")
        async def add_process_time_header(request: fastapi_request, call_next):
            self.before_request()
            response = await call_next(request)
            self.add_fastapi_metrics(request, response)
            return response
        

    def register_metrics(self):
        """
        Register PrometheusMiddleware to collect all default metrics for all endpoint.
        With exception of condition is meet with excluded list.
        """
        if isinstance(self.app, Flask) and not self._is_registered:
            # self.add_flask_middleware()
            self._is_registered = True
        elif isinstance(self.app, FastAPI) and not self._is_registered:
            self.add_fastapi_middleware()
            self._is_registered = True



import os


_DEFAULT_BUCKETS_BACKUP = [0.05, 0.1, 0.5, 0.9, 0.95]
# buckets in milliseconds #
_DEFAULT_BUCKETS = [0, 5, 10, 25, 50, 75, 100, 250, 500, 750, 1000, 2500, 5000, 7500, 10000,]
_DEFAULT_LABELS = ["http_request_method", "http_route", "http_response_status_code", "service_name", "service_namespace"]

UNDEFINED = "UNDEFINED"

NAMESPACE = os.getenv("SERVICE_NAME_PREFIX", UNDEFINED)
SERVICENAME = os.getenv("SERVICE_NAME", UNDEFINED)
SERVICEVERSION = os.getenv("SERVICE_VERSION", "1.0.0")
EXCLUDE_URLS = os.getenv("OTEL_PYTHON_EXCLUDED_URLS", "")
DICT_REQURIED_LABELS = {"service_namespace": NAMESPACE, "service_name": SERVICENAME}
SECONDTOMS = 1000



import prometheus_client

# from prometheus_client import Counter,Summary,Histogram,Gauge

from .config import _DEFAULT_BUCKETS
from .config import NAMESPACE, SERVICENAME, DICT_REQURIED_LABELS
from .custom_metrics_wrapper import (
    CounterWrapper,
    SummaryWrapper,
    HistogramWrapper,
    GaugeWrapper,
)
from ..utils.bcolor import Bcolors


class MetricCollector:
    def __init__(self) -> None:
        pass

    @staticmethod
    def _add_default_label(input_labels: list):
        for expected_label in DICT_REQURIED_LABELS:
            if expected_label not in input_labels:
                input_labels.append(expected_label)
        return input_labels

    @classmethod
    def gauge(
        cls,
        name: str,
        documentation: str,
        labels: list = [],
        multiprocess_mode: str = "all",
    ):
        """
        Gauge usage is like counter but can be decrease(.dec()) as well as increase(.inc())

        :param name: name of the gauge.
        :type name: str

        :param documentation: description of this metric.
        :type documentation: str

        :param labels: (Optional)labels of this gauge.
        :type labels: tuple

        :param multiprocess_mode: (Optional) when using multiprocess change to multiprocess_mode='livesum'
        :type multiprocess_mode: str

        :return: gauge object
        :rtype: prometheus_client.Counter Object
        """
        labels = cls._add_default_label(labels)
        metric_guage = prometheus_client.Gauge(
            name, documentation, labelnames=labels, multiprocess_mode=multiprocess_mode
        )
        warpped_guage = GaugeWrapper(metric_guage)
        return warpped_guage

    @classmethod
    def histogram(
        cls,
        name: str,
        documentation: str,
        labels: list = [],
        buckets: list = _DEFAULT_BUCKETS,
    ):
        """
        :param name: name of the histogram.
        :type name: str

        :param documentation: description of this metric.
        :type documentation: str

        :param labels: (Optional)labels of this histogram.
        :type labels: tuple

        :param buckets: (Optional)buckets of this histogram Ex buckets=[1,2,3,5,10].
        :type buckets: list

        :return: histogram object
        :rtype: prometheus_client.Counter Object
        """
        labels = cls._add_default_label(labels)
        metric_histogram = prometheus_client.Histogram(
            name,
            documentation,
            labelnames=labels,
            buckets=buckets,
        )
        warpped_histogram = HistogramWrapper(metric_histogram)
        return warpped_histogram

    @classmethod
    def summary(cls, name: str, documentation: str, labels: list = []):
        """
        A Summary tracks the size and number of events.

        :param name: name of the summary.
        :type name: str

        :param documentation: description of this metric.
        :type documentation: str

        :param labels: (Optional)labels of this summary.
        :type labels: tuple

        :return: summary object
        :rtype: prometheus_client.Counter Object

        Example use cases for Summary:

        Response latency
        Request size
        """
        labels = cls._add_default_label(labels)
        metric_summary = prometheus_client.Summary(
            name, documentation, labelnames=labels
        )
        warpped_summary = SummaryWrapper(metric_summary)

        return warpped_summary

    @classmethod
    def counter(cls, name: str, documentation: str, labels: list = []):
        """
        create counter metrics

        :param name: name of the counter.
        :type name: str

        :param documentation: description of this metric.
        :type documentation: str

        :param labels: (Optional)labels of this counter.
        :type labels: tuple

        :return: counter object
        :rtype: prometheus_client.Counter Object

        Example

        c = counter('my_failures_total', 'Description of counter')
        c.inc() # Increment by 1
        c.inc(1.6) # Increment by given value

        There are utilities to count exceptions raised:
        @c.count_exceptions() def f():
            pass

        with c.count_exceptions():
            pass

        # Count only one type of exception with c.count_exceptions(ValueError):
            pass
        """

        labels = cls._add_default_label(labels)
        metric_counter = prometheus_client.Counter(
            name,
            documentation,
            labelnames=labels,
        )
        warpped_counter = CounterWrapper(metric_counter)

        return warpped_counter



import unittest
import math
from src.thinknet_observer import Calculator


class CalculatorTestCase(unittest.TestCase):
    def setUp(self):
        self.calculator = Calculator()

    def test_natural_number(self):
        """Test natural number 5 multiplied by 2"""

        # 5 multiplied by 2 return 10
        result = self.calculator.multiply(5, 2)
        self.assertEqual(result, 10)

    def test_integer_number1(self):
        """Test integer number -7 multiplied by 2"""

        # -7 multiplied by 2 return -14
        result = self.calculator.multiply(-7, 2)
        self.assertEqual(result, -14)

    def test_integer_number2(self):
        """Test integer number -7 multiplied by 2"""

        # -7 multiplied by 2 return -14
        result = self.calculator.add(-7, 2)
        self.assertEqual(result, -5)

    def test_zero(self):
        """Test 0 multiplied by 2"""

        # 0 multiplied by 2 return 0
        result = self.calculator.multiply(2, 0)
        self.assertEqual(result, 0)

    def test_rational_number(self):
        """Test rational number 6/17 multiplied by 2"""

        # 6/17 multiplied by 2 return (6/17) * 2
        result = self.calculator.multiply(6 / 17, 2)
        self.assertEqual(result, (6 / 17) * 2)

    def test_real_number(self):
        """Test real number PI multiplied by 2"""

        # PI multiplied by 2 return 2PI
        result = self.calculator.multiply(math.pi, 2)
        self.assertEqual(result, math.pi * 2)


if __name__ == "__main__":
    unittest.main()



import time

from flask import request, Response

from ..metrics.metrics import PrometheusMiddleware
from ..utils.http import HttpContext
from ..loggers.loggers import TNAccessLogger

from . import get_resource_info
        

class FlaskObserverMiddleware:

    def __init__(self, app, resource=None, is_multiprocessing: bool=False, excluded_urls:list=[]) -> None:
        self._is_multiprocessing = is_multiprocessing
        self._excluded_urls = excluded_urls
        self._resource = resource

        if app is not None:
            self.init_app(app)


    def _setup_metrics (self, app,):
        namespace, service_name = get_resource_info(self._resource)

        self._metrics = PrometheusMiddleware(app, namespace, service_name, 
                                            is_multiprocessing=self._is_multiprocessing)
        self._metrics.add_exclude(self._excluded_urls)
        self._metrics.register_flask_metrics()


    def init_app(self, app, ):
        self.app = app
        self._request = request

        self._setup_metrics(app)
        self._access_log = TNAccessLogger(self._resource)

        self.app.before_request(self._before_request)
        self.app.after_request(self._after_response)


    def _before_request(self,):
        self.tick = time.time()


    def _after_response(self, response: Response):
        duration  = time.time() - self.tick

        http_context = HttpContext(self._request.method, self._request.path, self._request.referrer, 
                                self._request.remote_addr, duration, self._request.user_agent, 
                                self._request.url_rule, response.status_code)
        
        self._access_log.logging(http_context)
        self._metrics.add_flask_metrics(http_context)
        
        
        return response
